from blockchain import Blockchain
from simulate import Event
from utils import sample_exponential, gen_txn_id, get_txn_id
import random

#TODO(SM): Need to bind each transaction with a time of arrival at a node
    

class Node:
    def __init__(self, sim, ID, bandwidth_type, cpu_type, Tk, T_tx, num_nodes, pij):
        # Initialize node attributes
        self.sim = sim
        self.balance = 0
        self.nodeId = ID
        self.cpuType = cpu_type
        self.bandwidth_type = bandwidth_type
        self.Tk = Tk
        self.I = 600
        self.T_tx = T_tx
        self.num_nodes = num_nodes
        self.pij = pij

        self.neighbors = []
        self.blockchain = Blockchain()

    def init_simulation(self):
        """
            Initialize the simulation for generatinv transactions
            TODO: Schedule Block creation event
        """
        # Initialize simulation
        # schedule the first transaction
        self.sim.push_event(Event(self.sim.curr_time + sample_exponential(self.T_tx),\
                                  self._event_generate_transaction))
        
        self.sim.push_event(Event(self.sim.curr_time + sample_exponential(self.Tk),\
                                  self._event_generate_block, self.blockchain.create_new_block_for_pow()))
    
    def get_latency(self, node, data):
        """
            Get the latency to send the transaction to the given node
        """
        
        # check if the bandwidth type is slow for either of the nodes
        if(self.bandwidth_type == 1 and node.bandwidth_type == 1):
            cij = 100
        else:
            cij = 5
        cij *= 1e6

        dij = sample_exponential((96*1e3)/cij)
        latency = dij + self.pij

        if(type(data) == str):  # transaction
            m = 1024
        else:
            # TODO(AG): latency for block
            m = 1024
        
        return latency + (float(m)/cij) 
      
    def _event_generate_transaction(self, args):
        """
            Generate a transaction for the given peer
            Format: "TxnID: IDx pays IDy C coins"
        """  

        # Select a random destination node and amount
        dest = self.nodeId
        while dest == self.nodeId:
            dest = random.randint(0, self.num_nodes)
        amount = random.randint(1, self.balance + 10)

        # Create a transaction and it to current nodes pending transactions
        txn = str(gen_txn_id()) + ": " + str(self.nodeId) + " pays "\
            + str(dest) + " " + str(amount) + " coins"
        print(self.sim.curr_time, self.nodeId, "Generated a txn:", txn)

        self.pending_transactions.append(txn)

        # schedule the next transaction
        self.sim.push_event(Event(self.sim.curr_time + sample_exponential(self.T_tx),\
                                  self._event_generate_transaction))
        
        # broadcast the transaction to the neighbors
        self.broadcast_transaction(txn)
        pass

    def _event_receive_transaction(self, txn):
        """
            Event to receive a transaction from a neighbor
        """
        txn = txn[0]
        print(self.sim.curr_time, self.nodeId, "Received transaction", get_txn_id(txn))
        # Receive the transaction and add it to the pending transactions
        # If the transaction is new, add it and broadcast to the neighbors
        if self.blockchain.receive_txn(txn) == 1:
            # Broadcast the received transaction to the neighbors immediately
            self.broadcast_transaction(txn)
        else:
            print(self.sim.curr_time, self.nodeId, "Already exist, discarding ", get_txn_id(txn))
        return
    
    def broadcast_transaction(self, txn):
        """
            Broadcast the transaction to the neighbors
        """
        print(self.sim.curr_time, self.nodeId, "Broadcasting txn", get_txn_id(txn))
        for neighbor in self.neighbors:
            print(self.sim.curr_time, self.nodeId, "Sending txn", get_txn_id(txn),\
                  "to", neighbor.nodeId, self.sim.curr_time\
                    + self.get_latency(neighbor, txn))
            
            self.sim.push_event(Event(self.sim.curr_time + self.get_latency(neighbor, txn),\
                                      neighbor._event_receive_transaction, txn))
        return
  

        
    def _event_generate_block(self, blk):
        """
            Event to perform create a block and perform PoW and broadcast it
        """
        # discard the event if the time is not right
        # this means this event has been overriden
        if self.sim.curr_time != self.blockchain.next_pow_completion_time:
            return 
        
        # Add the block it to the blockchain and broadcast it to the neighbors
        self.blockchain.add_new_mined_block(blk)
        self.broadcast_block(blk)

        # Schedule the next block creation event
        new_blk = self.blockchain.create_new_block_for_pow()
        self.sim.push_event(Event(self.sim.curr_time + sample_exponential(self.Tk),\
                                  self._event_generate_block, new_blk))
        return
    
    def _event_receive_block(self, blk):
        """
            Event to receive a block from a neighbor
            Receive the block and validate it's transactions / hash
            If valid, check if it is the longest chain
            If it is the longest chain, add this block to the blockchain
            Schedule the next block creation event
            If not valid ignore
        """

        if self.blockchain.alread_exist(blk) or self.blockchain.verify_blk(blk) == 0:
            return
        
        self.blockchain.add_new_mined_block(blk)
        print(self.sim.curr_time, self.nodeId, "Received block", blk.blkId)
        chain_length = self.blockchain.get_chain_length(blk)
        if chain_length > self.blockchain.len_longest_chain:
            self.blockchain.len_longest_chain = self.blockchain.get_chain_length(blk)
            self.blockchain.head = blk.blkId
            self.sim.push_event(Event(self.sim.curr_time + sample_exponential(self.Tk),\
                                        self._event_generate_block, self.blockchain.create_new_block_for_pow()))

        self.broadcast_block(blk)

    def broadcast_block(self, blk):
        """
            Broadcast the block to the neighbors
        """
        for neighbor in self.neighbors:
            self.sim.push_event(Event(self.sim.curr_time + self.get_latency(neighbor, blk),\
                                      neighbor._event_receive_blk, blk))
        return